#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/Command.h>
#include <Xm/FileSB.h>
#include <Xm/MessageB.h>
#include <X11/Xlib.h>
#include <X11/cursorfont.h>
#include "motif.h"

#define declareCB(name) void name(Widget w, XtPointer client_data, XmAnyCallbackStruct *cbs);
#define declareCBstru(name,stru) void name(Widget w, XtPointer client_data, stru cbs);
#define declareCBcli(name,cli) void name(Widget w, cli client_data,  XmAnyCallbackStruct *cbs);

#define defineRegistration(func, wid) \
Widget wid = NULL; \
void func(Widget w, \
		  XtPointer client_data, \
		  XmAnyCallbackStruct *cbs) \
{ \
	wid = w; \
}


declareCB(registerTextWindowCB);
declareCB(registerOpenMenuCB);
declareCB(registerSearchCommandCB);
declareCB(registerRangeCommandCB);
declareCB(registerStatusLineCB);
declareCB(registerBookmarkStatusLineCB);
declareCB(registerGotoDialogCB);
declareCB(registerBookSelectCB);
declareCB(registerGotoCommandCB);
declareCB(registerBookmarkListCB);
declareCB(registerMenu_BookmarkCB);
declareCB(registerBookmarkNameEntryCB);
declareCB(registerBookmarkVerseEntryCB);
declareCBcli(openVersionCB, char *);
declareCBstru(bookOkCB, XmSelectionBoxCallbackStruct *);
declareCBstru(bookApplyCB, XmSelectionBoxCallbackStruct *);
declareCBstru(bookCancelCB, XmSelectionBoxCallbackStruct *);
declareCBstru(exitCB, XmSelectionBoxCallbackStruct *);
declareCB(saveAsCB);
declareCB(printerCB);
declareCB(gotoCB);
declareCB(bookListCB);
declareCB(bookmarkCB);
declareCB(searchCB);
declareCB(registerSearchDialogCB);
declareCB(search_nextCB);
declareCB(search_prevCB);
declareCB(search_firstCB);
declareCB(search_lastCB);
declareCB(search_cancelCB);
declareCB(search_helpCB);
declareCB(goto_okCB);
declareCB(goto_cancelCB);
declareCB(goto_helpCB);
declareCB(bookmark_openCB);
declareCB(bookmark_loadCB);
declareCB(bookmark_saveCB);
declareCB(bookmark_saveasCB);
declareCB(bookmark_clearCB);
declareCBstru(bookmark_gotoCB, XmListCallbackStruct *);
declareCB(bookmark_cancelCB);
declareCB(bookmark_popdownCB);
declareCB(bookmark_printCB);
declareCB(bookmark_addCB);
declareCB(bookmark_updateCB);
declareCB(bookmark_deleteCB);
declareCBstru(bookmark_selectCB, XmListCallbackStruct *);
declareCBstru(bookmark_positionCB, XmListCallbackStruct *);

#define regMem(name) { #name, (caddr_t)name }

static MRMRegisterArg register_vector[] =
	{
	regMem(registerTextWindowCB),
	regMem(registerOpenMenuCB),
	regMem(registerStatusLineCB),
	regMem(registerBookmarkStatusLineCB),
	regMem(registerBookSelectCB),
	regMem(registerGotoDialogCB),
	regMem(registerSearchCommandCB),
	regMem(registerRangeCommandCB),
	regMem(registerGotoCommandCB),
	regMem(registerBookmarkListCB),
	regMem(registerMenu_BookmarkCB),
	regMem(registerBookmarkNameEntryCB),
	regMem(registerBookmarkVerseEntryCB),
	regMem(registerMenu_BookmarkCB),
	regMem(bookOkCB),
	regMem(bookApplyCB),
	regMem(bookCancelCB),
	regMem(exitCB),
	regMem(saveAsCB),
	regMem(printerCB),
	regMem(gotoCB),
	regMem(bookListCB),
	regMem(bookmarkCB),
	regMem(searchCB),
	regMem(registerSearchDialogCB),
	regMem(search_nextCB),
	regMem(search_prevCB),
	regMem(search_firstCB),
	regMem(search_lastCB),
	regMem(search_cancelCB),
	regMem(search_helpCB),
	regMem(goto_okCB),
	regMem(goto_cancelCB),
	regMem(goto_helpCB),
	regMem(bookmark_openCB),
	regMem(bookmark_loadCB),
	regMem(bookmark_saveCB),
	regMem(bookmark_saveasCB),
	regMem(bookmark_clearCB),
	regMem(bookmark_gotoCB),
	regMem(bookmark_cancelCB),
	regMem(bookmark_popdownCB),
	regMem(bookmark_printCB),
	regMem(bookmark_addCB),
	regMem(bookmark_updateCB),
	regMem(bookmark_deleteCB),
	regMem(bookmark_selectCB),
	regMem(bookmark_positionCB),
	};

Widget openWidget = NULL;
Widget toplevel = NULL;
Widget bookmarkWidget = NULL;
Widget bookmarkFileWidget = NULL;
char * error_open = "Open a book first";
char * error_version = "Open a version first";
BookmarkNum bookmarkLastSelected = -1;
static Cursor		watchCursor = None;
static Cursor		stdCursor = None;

defineRegistration(registerTextWindowCB, textWidget);
defineRegistration(registerStatusLineCB, statusLineWidget);
defineRegistration(registerBookmarkStatusLineCB, bookmarkStatusLineWidget);
defineRegistration(registerSearchCommandCB, searchCommandWidget);
defineRegistration(registerRangeCommandCB, rangeCommandWidget);
defineRegistration(registerBookSelectCB, bookSelectWidget);
defineRegistration(registerSearchDialogCB, searchDialogWidget);
defineRegistration(registerGotoDialogCB, gotoDialogWidget);
defineRegistration(registerGotoCommandCB, gotoCommandWidget);
defineRegistration(registerBookmarkListCB, bookmarkListWidget);
defineRegistration(registerMenu_BookmarkCB, menu_bookmarkWidget);
defineRegistration(registerBookmarkNameEntryCB, bookmarkNameEntryWidget);
defineRegistration(registerBookmarkVerseEntryCB, bookmarkVerseEntryWidget);

#define register_vector_length \
((sizeof register_vector) / (sizeof register_vector[0]))
MrmHierarchy hier;


static printmessage(char * mesg, Widget w)
	{
		if (w == NULL)
			fprintf(stderr, "%s\n", mesg);
		else
			{
			XtVaSetValues(w, XmNvalue, mesg, NULL);
			XFlush(XtDisplay(w));
			}
	}

static void domessage(char * mesg, Widget w)
	{
	if (mesg == NULL)
		printmessage("", w);
	else
		printmessage(mesg, w);
	}

static void message(char * mesg)
	{
	domessage(mesg, statusLineWidget);
	}

static void bookmarkMessage(char * mesg)
	{
	domessage(mesg, bookmarkStatusLineWidget);
	}

void destroyCB(Widget w,
			   XtPointer client_data,
			   XmAnyCallbackStruct *cbs)
	{
	XtDestroyWidget(w);
	}

void errorDialog(char * mesg)
	{
	XmString cs;
	Arg args[10];
	int i = 0;
	Widget dialog, cb;
	
	cs = XmStringCreateSimple(mesg);
	XtSetArg(args[i], XmNmessageString, cs); i++;
	XtSetArg(args[i], XmNdeleteResponse, XmDESTROY); i++;
	dialog = XmCreateErrorDialog(toplevel, "error", args, i);
	XmStringFree(cs);
	XtVaGetValues(dialog, XmNcancelButton, &cb, NULL);
	XtDestroyWidget(cb);
	XtAddCallback(dialog, XmNokCallback, (XtCallbackProc)destroyCB, NULL);
	XtManageChild(dialog);
	}

int runmotif(MotifContext *mcon)
	{
	XtAppMainLoop(mcon->appContext);
	return 0;
	}

String fallbacks[] = 
	{
	"*bib*viewer*columns: 80",
	"*bib*viewer*rows: 20",
	NULL
	};

static busy(Widget w)
	{
	XDefineCursor(XtDisplay(w), XtWindow(w), watchCursor);
	XFlush(XtDisplay(w));
	}

static notbusy(Widget w)
	{
	XDefineCursor(XtDisplay(w), XtWindow(w), stdCursor);
	}

static setCursor(Widget w)
	{
	notbusy(w);
	}

extern Widget doFetch( MrmHierarchy hierarchy_id , String index , Widget parent , Widget *w_return , MrmType *class_return )
	{
	Widget rtn;
	rtn = XtCreatePopupShell(index, topLevelShellWidgetClass, parent, NULL, 0); 
 	if (MrmFetchWidget(hierarchy_id, index, rtn, w_return, class_return) != MrmSUCCESS)
		message("cannot fetch");
	if (*w_return != NULL)
		{
		if (watchCursor == None)
			watchCursor = XCreateFontCursor(XtDisplay(toplevel), XC_watch);
		if (stdCursor == None)
			stdCursor = XCreateFontCursor(XtDisplay(toplevel), XC_top_left_arrow);
		XtManageChild(*w_return);
		setCursor(toplevel);
		XtPopup(rtn, XtGrabNone);
		}
	else
		message("Cannot fetch widget");
	return rtn;
	}

void setTitle(Widget w, char * title)
{
	XtVaSetValues(w, XmNtitle, title,
				  NULL);
}

MotifContext * loadmotif(int argc, char **argv)
	{
	Widget aWidget;
	static char *uid_files[] = {"bib.uid"};
	MrmType aType;
	MotifContext * mcon;

	if (getenv("UIDPATH") ==  NULL)
		{
		char value[100];
		cpystr(value, "UIDPATH=./%U:", bhome, "/uid/%U", NULL);
		putenv(value);
		}
	mcon = (MotifContext *)malloc(sizeof (MotifContext));
	toplevel = XtVaAppInitialize(&mcon->appContext, "Bib", NULL, 0, &argc, argv, fallbacks, NULL);
	MrmInitialize();
	if (MrmOpenHierarchy(1, uid_files, NULL, &hier) != MrmSUCCESS)
		{
		fprintf(stderr, "Cannot open Hierarchy\n");
		return NULL;
		}
	MrmRegisterNames(register_vector, register_vector_length);
	if (MrmFetchWidget(hier, "bib", toplevel, &aWidget, &aType) != MrmSUCCESS)
		{
		fprintf(stderr, "Cannot fetch top Widget\n");
		return NULL;
		}
	XtManageChild(aWidget);
	XtRealizeWidget(toplevel);
	watchCursor = XCreateFontCursor(XtDisplay(toplevel), XC_watch);
	stdCursor = XCreateFontCursor(XtDisplay(toplevel), XC_top_left_arrow);
	mcon->context.run = (int (*)(Context *))runmotif;
	return mcon;
	}

XmString * vecToXmStringVec(char ** vec, int size)
	{
	XmString * rtn, * sptr;
	rtn = sptr = (XmString *)malloc(sizeof(XmString) * (size + 1));
	for (; *vec != NULL; vec++, sptr++)
		*sptr = XmStringCreateSimple(*vec);
	*sptr = NULL;
	return rtn;
	}

void freeXmStringVec(XmString * vec)
	{
	XmString * ptr;
	ptr = vec;
	while (*ptr != NULL)
		XmStringFree(*ptr++);
	free(vec);
	}

void vecToMenu(Widget w, XtCallbackProc p, char ** vec, int size)
	{
	XmString str;
	Widget button;
	for (; *vec != NULL; vec++)
		{
		str = XmStringCreateSimple(*vec);
		button = XtVaCreateManagedWidget("button", xmPushButtonWidgetClass, w,
										 XmNlabelString, str,
										 NULL);
		XtAddCallback(button, XmNactivateCallback, p, strdup(*vec));
		XmStringFree(str);
/*		XtManageChild(button); */
		}
	}

void registerOpenMenuCB(Widget w,
						XtPointer client_data,
						XmAnyCallbackStruct *cbs)
	{
	char ** vec;
	int num;
	char filename[PATHSIZE+1];

	openWidget = w;
	
	vec = getVersionListSVec(&num);
	if (vec == NULL)
		message("Contents file not found");
	else
		{
		vecToMenu(w, (XtCallbackProc)openVersionCB, vec, num);
		freevector(vec);
		}
	}


void populateSelectionFromVec(Widget w, char ** vec, int num)
{
	FILE * file;
	char buf[500];
	XmString * svec;

	svec = vecToXmStringVec(vec, num);
	XtVaSetValues(w, XmNlistItemCount, num,
				  XmNlistItems, svec,
				  NULL);
	freeXmStringVec(svec);
}

Bool popupWindow(MrmHierarchy hier, char * name, Widget parent, Widget * widget, MrmType * type)
	{
	Bool rtn = YES;
	if (MrmFetchWidget(hier, name, parent, widget, type) != MrmSUCCESS)
		{
		message("cannot fetch");
		rtn = NO;
		}
	return rtn;
	}


void openVersionCB(Widget w,
			 char * client_data,
			 XmAnyCallbackStruct *cbs)
	{
	cVersion = loadVersion(client_data);
	setTitle(toplevel, cVersion->title);
	}

void gotoCB(Widget w,
			 XtPointer client_data,
			 XmAnyCallbackStruct *cbs)
	{
	Widget sel;
	MrmType mtype;
	
	doFetch(hier, "goto_dialog", w, &sel, &mtype);
	}

void bookListCB(Widget w,
				XtPointer client_data,
				XmAnyCallbackStruct *cbs)
	{
	MrmType mtype;
	Widget sel;
	
	if (cVersion == NULL)
		errorDialog(error_version);
	else if (doFetch(hier, "book_selection", w, &sel, &mtype))
		{
		int num;
		char ** vec;
		XmString head;
		
		vec = getBookContentsSVec(cVersion->title, &num);
		if (vec == NULL)
			errorDialog("Cannot find Contents file");
		else
			{
			populateSelectionFromVec(bookSelectWidget, vec, num);
			freevector(vec);
			
			head = XmStringCreateSimple(cVersion->title);
			XtVaSetValues(bookSelectWidget, XmNdialogTitle, head,
						  NULL);
			XmStringFree(head);
			XtManageChild(sel);
			setCursor(sel);
			}
		}
	}

void bookmark_cancelCB(Widget w, XtPointer client_data,
					   XmAnyCallbackStruct *cbs) { XtDestroyWidget(XtParent(bookmarkWidget));
					   XtVaSetValues(menu_bookmarkWidget, XmNsensitive, TRUE, NULL);
					   bookmarkWidget = NULL; }

void showBookmarks(Widget wList, BookmarkList * blist)
	{
	XmStringTable table;
	int count;
	char buf[1024+1];
	Bookmark * bmark;
	char rbuf[1024+1];
	
	if (blist == NULL)
		XtVaSetValues(wList, XmNitemCount, 0, NULL);
	else
		{
		table = (XmStringTable)malloc(blist->num * sizeof(XmString));
		for (count = 0; count < blist->num; count++)
			{
			bmark = &blist->bookmarks[count];
			sprintf(buf, "%-20s %s", bmark->title, rangeToBuf(bmark->range, rbuf));
			table[count] = XmStringCreateSimple(buf);
			}
		XtVaSetValues(wList, XmNitems, table, XmNitemCount, count, NULL);
		}
	}

void bookmarkCB(Widget w,
				XtPointer client_data,
				XmAnyCallbackStruct *cbs)
	{
	if (bookmarkWidget == NULL)
		{
		MrmType mtype;
		doFetch(hier, "bookmark_dialog", w, &bookmarkWidget, &mtype);
		XtAddCallback(XtParent(bookmarkWidget), XmNpopdownCallback, (XtCallbackProc)bookmark_cancelCB, NULL);
		/* This call causes an Xt error. WHY????????? Not any more! */
		}
	if (bookmarkWidget != NULL)
		{
		XtVaSetValues(menu_bookmarkWidget, XmNsensitive, FALSE, NULL);
		if (bmarklist == NULL)
			bmarklist = bookmark_create(0);
		showBookmarks(bookmarkListWidget, bmarklist);
		}
	}

void searchCB(Widget w,
			  XtPointer client_data,
			  XmAnyCallbackStruct *cbs)
	{
	Widget sel;
	MrmType mtype;
	doFetch(hier, "search_dialog", w, &sel, &mtype);
	}

void bookCancelCB(Widget w,
			  XtPointer client_data,
			  XmSelectionBoxCallbackStruct *cbs)
	{
	XtDestroyWidget(XtParent(w));
	}

void bookApplyCB(Widget w,
			  XtPointer client_data,
			  XmSelectionBoxCallbackStruct *cbs)
	{
	char *value;
	char *bookContents, *headvalue;
	Book * book;
	XmString head;
	struct stat sstat;
	char filename[PATHSIZE];
	int fd;
	
	XtVaGetValues(XtParent(w), XmNdialogTitle, &head, NULL);
	XmStringGetLtoR(head, XmSTRING_DEFAULT_CHARSET, &headvalue);
	XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &value);
	cVersion = loadBookVersion(headvalue, value);
	XtVaSetValues(textWidget, XmNvalue, cVersion->cBook->data, NULL);
	setTitle(toplevel, cVersion->cBook->title);
	}

void bookOkCB(Widget w,
			 XtPointer client_data,
				 XmSelectionBoxCallbackStruct *cbs)
	{
	bookApplyCB(w, client_data, cbs);
	XtDestroyWidget(XtParent(XtParent(w)));
	}

void exitCB(Widget w,
			XtPointer client_data,
			XmSelectionBoxCallbackStruct *cbs)
	{
	exit(0);
	}
void saveAsCB(Widget w,
			  XtPointer client_data,
			  XmAnyCallbackStruct *cbs)
	{
	Widget sel;
	MrmType mtype;
	doFetch(hier, "save_as", w, &sel, &mtype);
	}

void printerCB(Widget w,
			   XtPointer client_data,
			   XmAnyCallbackStruct *cbs)
	{
	Widget sel;
	MrmType mtype;
	doFetch(hier, "printer_dialog", w, &sel, &mtype);
	}

static void showVerse(Widget w, Book * book, AbsVerse ver)
{
	message(NULL);
	XmTextSetCursorPosition(w, book->verses[ver]);
	XmTextSetCursorPosition(w, endOfVerse(book, ver));
}

char * searching = "Searching...";

void searchForward(Pos pos)
{
	int ver;
	char * expression;
	char * srange;
	Range * range;
	ErrorBuf errbuf;
	Bool nBook;
	int eaten;

	if (cVersion == NULL)
		errorDialog(error_open);
	else
	{
		message(searching);
		busy(searchDialogWidget);
		expression = XmTextGetString(XmCommandGetChild(searchCommandWidget, XmDIALOG_COMMAND_TEXT));
		srange = XmTextGetString(XmCommandGetChild(rangeCommandWidget, XmDIALOG_COMMAND_TEXT));
		range = parseVerse(srange, &eaten);
		ver = nextSearchVerse(cVersion->cBook, pos);
		ver = searchNextVersion(cVersion, ver, range, expression, errbuf, NO, &nBook);
		if (0 <= ver)
		{
			if (nBook)
				{
				XtVaSetValues(textWidget, XmNvalue, cVersion->cBook->data, NULL);
				setTitle(toplevel, cVersion->cBook->title);
				}
			showVerse(textWidget, cVersion->cBook, ver);
			message(NULL);
		}
		else
			message(errbuf);
		notbusy(searchDialogWidget);
	}
}

void searchBack(Pos pos)
{
	AbsVerse ver;
	char * expression;
	char * srange;
	Range * range;
	ErrorBuf errbuf;
	Bool nBook;
	int eaten;

	if (cVersion == NULL)
		errorDialog(error_open);
	else
	{
		message(searching);
		busy(searchDialogWidget);
		expression = XmTextGetString(XmCommandGetChild(searchCommandWidget, XmDIALOG_COMMAND_TEXT));
		srange = XmTextGetString(XmCommandGetChild(rangeCommandWidget, XmDIALOG_COMMAND_TEXT));
		range = parseVerse(srange, &eaten);
		ver = prevSearchVerse(cVersion->cBook, pos);
		ver = searchNextVersion(cVersion, ver, range, expression, errbuf, YES, &nBook);
		if (0 <= ver)
		{
			if (nBook)
				{
				XtVaSetValues(textWidget, XmNvalue, cVersion->cBook->data, NULL);
				setTitle(toplevel, cVersion->cBook->title);
				}
			showVerse(textWidget, cVersion->cBook, ver);
			message(NULL);
		}
		else
			message(errbuf);
		notbusy(searchDialogWidget);
	}
}

void search_cancelCB(Widget w,
				XtPointer client_data,
				XmAnyCallbackStruct *cbs)
{
	XtDestroyWidget(XtParent(searchDialogWidget));
}

void search_helpCB(Widget w,
			  XtPointer client_data,
			  XmAnyCallbackStruct *cbs)
	{
	}


void search_firstCB(Widget w,
					XtPointer client_data,
					XmAnyCallbackStruct *cbs)
	{
	if (cVersion == NULL)
		errorDialog(error_open);
	else
		{
		Bool searchAll = YES;
		busy(searchDialogWidget);
		if (searchAll)
			{
			loadBookAsCurrent(cVersion, cVersion->books[0].title);
			XtVaSetValues(textWidget, XmNvalue, cVersion->cBook->data, NULL);
			setTitle(toplevel, cVersion->cBook->title);
			}
		searchForward(0);
		notbusy(searchDialogWidget);
		}
	}

void search_lastCB(Widget w,
				   XtPointer client_data,
				   XmAnyCallbackStruct *cbs)
	{
	if (cVersion == NULL)
		errorDialog(error_open);
	else
		{

		Bool searchAll = YES;
		busy(searchDialogWidget);
		if (searchAll)
			{
			loadBookAsCurrent(cVersion, cVersion->books[cVersion->numBooks-1].title);
			XtVaSetValues(textWidget, XmNvalue, cVersion->cBook->data, NULL);
			setTitle(toplevel, cVersion->cBook->title);
			}
		searchBack(cVersion->cBook->size - 1);
		notbusy(searchDialogWidget);
		}
	}

void search_nextCB(Widget w,
				   XtPointer client_data,
				   XmAnyCallbackStruct *cbs)
	{
	searchForward(XmTextGetCursorPosition(textWidget));
	}

void search_prevCB(Widget w,
				   XtPointer client_data,
				   XmAnyCallbackStruct *cbs)
	{
	searchBack(XmTextGetCursorPosition(textWidget));
	}


void goto_okCB(Widget w,
			  XtPointer client_data,
			  XmAnyCallbackStruct *cbs)
{
	AbsVerse ver;
	Bool nBook;
	char * verse;

	if (cVersion == NULL)
		errorDialog(error_open);
	else
	{
		busy(gotoDialogWidget);
		verse = XmTextGetString(XmCommandGetChild(gotoCommandWidget, XmDIALOG_COMMAND_TEXT));
		ver = gotoVerseString(cVersion, verse, &nBook);
		if (ver == noVerseNum)
			message("invalid verse");
		else
			{
			if (nBook)
				{
				XtVaSetValues(textWidget, XmNvalue, cVersion->cBook->data, NULL);
				setTitle(toplevel, cVersion->cBook->title);
				}
			showVerse(textWidget, cVersion->cBook, ver);
			}
		notbusy(gotoDialogWidget);
	}
}
void goto_cancelCB(Widget w,
			  XtPointer client_data,
			  XmAnyCallbackStruct *cbs)
{
	XtDestroyWidget(XtParent(gotoDialogWidget));
}

void goto_helpCB(Widget w,
			  XtPointer client_data,
			  XmAnyCallbackStruct *cbs)
{
}

void fileselectCB(Widget w,
					 void (*client_data)(char *),
					 XmFileSelectionBoxCallbackStruct *cbs)
{
	char * filename;
	if (!XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename))
		return;
	client_data(filename);
	XtFree(filename);
	XtUnmanageChild(w);
}

void filecancelCB(Widget w,
					 void (*client_data)(char *),
					 XmFileSelectionBoxCallbackStruct *cbs)
{
	XtUnmanageChild(w);
}

Bool prompt_filename(Widget * w, void (*proc)(char *))
{
	if (*w == NULL)
	{
		*w = XmCreateFileSelectionDialog(toplevel, "fileselect", NULL, 0);
		XtAddCallback(*w, XmNokCallback, (XtCallbackProc)fileselectCB, proc);
		XtAddCallback(*w, XmNcancelCallback, (XtCallbackProc)filecancelCB, proc);
	}
	XtManageChild(*w);
}


void bookmark_open(char * filename)
{
	bmarklist = reopenBookmark(filename, bmarklist);
	showBookmarks(bookmarkListWidget, bmarklist);
}

void bookmark_load(char * filename)
{
	bmarklist = combinedLoadBookmark(filename, bmarklist);
	showBookmarks(bookmarkListWidget, bmarklist);
}

void bookmark_openCB(Widget w,
					 XtPointer client_data,
					 XmAnyCallbackStruct *cbs)
{
	prompt_filename(&bookmarkFileWidget, bookmark_open);
}
void bookmark_loadCB(Widget w,
					 XtPointer client_data,
					 XmAnyCallbackStruct *cbs)
{
	prompt_filename(&bookmarkFileWidget, bookmark_load);
}
void bookmark_saveCB(Widget w,
					 XtPointer client_data,
					 XmAnyCallbackStruct *cbs)
{
}
void bookmark_saveasCB(Widget w,
					   XtPointer client_data,
					   XmAnyCallbackStruct *cbs)
{
}

void bookmark_clearCB(Widget w,
					   XtPointer client_data,
					   XmAnyCallbackStruct *cbs)
{
	freeBookmarks(bmarklist);
	bmarklist = bookmark_create(0);
	showBookmarks(bookmarkListWidget, bmarklist);
}


void bookmark_gotoCB(Widget w,
					   XtPointer client_data,
					   XmListCallbackStruct *cbs)
	{
	AbsVerse ver;
	Bool nBook;
	Bookmark * mk;
	Verse * verse;
	
	if (cVersion == NULL)
		errorDialog(error_version);
	else if (0 <= bookmarkLastSelected)
		{
		Verse vse;
		mk = &bmarklist->bookmarks[bookmarkLastSelected];
		verse = &mk->range->items[0].from;
		
		memcpy(&vse, verse, sizeof(vse));
		normaliseFromVerse(&vse);
		ver = gotoVerse(cVersion, vse.title, vse.chap, vse.verse, &nBook);
		if (ver == noVerseNum)
			message("invalid verse");
		else
			{
			if (nBook)
				{
				XtVaSetValues(textWidget, XmNvalue, cVersion->cBook->data, NULL);
				setTitle(toplevel, cVersion->cBook->title);
				}
			showVerse(textWidget, cVersion->cBook, ver);
			}
	}
	else
		bookmarkMessage("Select a bookmark first");
}

void bookmark_printCB(Widget w,
					   XtPointer client_data,
					   XmAnyCallbackStruct *cbs)
{
	
}

void bookmark_addCB(Widget w,
					   XtPointer client_data,
					   XmAnyCallbackStruct *cbs)
	{
	Bookmark bmark;
	char * nm, * rng;
	int eaten;
	
	XtVaGetValues(bookmarkNameEntryWidget, XmNvalue, &nm, NULL);
	XtVaGetValues(bookmarkVerseEntryWidget, XmNvalue, &rng, NULL);
	bmark.title = strdup((char *)nm);
	bmark.range = parseVerse(rng, &eaten);
	bookmarkAdd(bmarklist, &bmark, bookmarkLastSelected);
	showBookmarks(bookmarkListWidget, bmarklist);
#ifdef NEVER
	XtVaSetValues(bookmarkListWidget, XmNtopItemPosition, bmarklist->num, NULL);
#endif // NEVER
	}

void bookmark_updateCB(Widget w,
					   XtPointer client_data,
					   XmAnyCallbackStruct *cbs)
{
	Bookmark bmark;
	char * nm, * rng;
	int eaten;
	
	XtVaGetValues(bookmarkNameEntryWidget, XmNvalue, &nm, NULL);
	XtVaGetValues(bookmarkVerseEntryWidget, XmNvalue, &rng, NULL);
	bmark.title = strdup((char *)nm);
	bmark.range = parseVerse(rng, &eaten);   
	bookmarkUpdate(bmarklist, &bmark, bookmarkLastSelected);
	showBookmarks(bookmarkListWidget, bmarklist);
	}

void bookmark_deleteCB(Widget w,
					   XtPointer client_data,
					   XmAnyCallbackStruct *cbs)
	{
	bookmarkDelete(bmarklist, bookmarkLastSelected);
	showBookmarks(bookmarkListWidget, bmarklist);
	}

void bookmark_positionCB(Widget w,
					   XtPointer client_data,
					   XmListCallbackStruct *cbs)
	{
	Bookmark * mk;
	char buf[RANGEBUF_SIZE+1];
	
	bookmarkLastSelected = cbs->item_position - 1;
	mk = &bmarklist->bookmarks[bookmarkLastSelected];
	XtVaSetValues(bookmarkNameEntryWidget, XmNvalue, mk->title, XmNcursorPosition, 0, NULL);
	XtVaSetValues(bookmarkVerseEntryWidget, XmNvalue, rangeToBuf(mk->range, buf), XmNcursorPosition, 0, NULL);
	
	}

void bookmark_selectCB(Widget w,
					   XtPointer client_data,
					   XmListCallbackStruct *cbs)
	{
	bookmarkLastSelected = cbs->item_position - 1;
	}

void bookmark_popdownCB(Widget w,
					   XtPointer client_data,
						XmAnyCallbackStruct *cbs)
	{
	XtVaSetValues(menu_bookmarkWidget, XmNsensitive, TRUE, NULL);
	}






